#include "Constant.h"
#include "math.h"
#include "MathStats.h"
#include <string>
#include <cstring>
#include <iostream>
#include "PreMeta.h"

String PreMeta::vcfInput = ""; //input annotated vcf file name here
int PreMeta::Xstart = 2699520;
int PreMeta::Xend = 154931044;
bool PreMeta::dosage =false;
bool PreMeta::genoFromPed =false;
bool PreMeta::genoFromVCF =false;
bool PreMeta::FounderFreq =false;
String PreMeta::outputFile = "";
String PreMeta::xLabel = "X";
int PreMeta::window = 1000000;

PreMeta::PreMeta(){}

PreMeta::~PreMeta()
{}

//TODO:sort markers by chr and pos
/*
   void PreMeta::PedSortMarkers()
   {
   String markerName;
   markerName.AddTokens(ped.markerNames[0]);
   String initial_chr = name[0]; 
   for(int m=0;m<ped.markerCount;m++)
   {


   }
   }
 */
void PreMeta::VCFSanityCheck()
{
   reader.open(vcfInput,header);
   String chrom;
   int position;
   while(reader.readRecord(record))
   {
      chrom = record.getChromStr();
      position = record.get1BasedPosition();
      break;
   }
   reader.close();

   if(chrom.SlowFind("chr",0)!=-1)
   {
      //chromosome in VCF has "chr" in it
      reader.open(vcfInput,header);
      String prev_chrom;
      int prev_pos;
      while(reader.readRecord(record))
      {
	 prev_chrom = chrom;
	 prev_pos = position;
	 chrom = record.getChromStr();
	 position = record.get1BasedPosition();
	 //compare here
	 if(chrom==prev_chrom)
	 {
	    if(position <= prev_pos)
	    {
	       error("FATAL ERROR! VCF file has to be sorted by chrom and pos.\n");
	       break;
	    }
	 }
	 else 
	 {
	    prev_chrom = prev_chrom.SubStr(3);
	    chrom = chrom.SubStr(3);
	    if(atoi(chrom.c_str()) < atoi(prev_chrom.c_str()))
	    {
	       error("FATAL ERROR! VCF file has to be sorted by chrom and pos.\n");
	       break;
	    }
	    else
	    {
	       continue;
	    }
	 }
	 reader.close();
      }
   }
   else 
   {
      reader.open(vcfInput,header);
      String prev_chrom;
      int prev_pos;
      while(reader.readRecord(record))
      {
	 prev_chrom = chrom;
	 prev_pos = position;
	 chrom = record.getChromStr();
	 position = record.get1BasedPosition();
	 //compare here
	 if(chrom==prev_chrom)
	 {
	    if(position <= prev_pos)
	    {
	       error("FATAL ERROR! VCF file has to be sorted by chrom and pos.\n");
	       break;
	    }
	 }
	 else if(atoi(chrom.c_str()) < atoi(prev_chrom.c_str())) 
	 {
	    error("FATAL ERROR! VCF file has to be sorted by chrom and pos.\n");
	    break;
	 }
	 else 
	 {
	    continue;
	 }
      }
      reader.close();
   }
}

double PreMeta::CalculateCov(FastFit & engine,FastTransform & trans, Pedigree & ped, Matrix & genotypeAll,Vector & sigma2,int m)
{
   double cov = 0.0;
   if(FastFit::unrelated)
   {    
      for(int p=0;p<trans.persons;p++)
      {     
	 cov += genotypeAll[p][0] * genotypeAll[p][m];
      }     
      cov *= engine.sigma2Hat;
   }    
   else 
   {    
      for(int p=0;p<trans.persons;p++)
      {     
	 cov += genotypeAll[p][0] * genotypeAll[p][m]/sigma2[p];
      }     
   }    
   cov /= trans.persons;
   return cov;
}

void PreMeta::Run(Pedigree & ped, FastTransform & trans,FastFit & engine, String & traitName)
{
   /*
      if(genoFromVCF)
      {
   //This is to make sure VCF file is sorted
   VCFSanityCheck();
   }
    */
   //TODO:if genotypes are from vcf, check if markers are sorted by chr and pos
   FILE * SCOREoutput;
   FILE * SCOREcov;
   String filename;

   if(outputFile != "")
      filename = outputFile + "." + traitName + ".singlevar.score.txt";
   else
      filename = traitName + ".singlevar.score.txt";

   SCOREoutput = fopen(filename,"wt");

   if(outputFile != "")
      filename = outputFile + "." + traitName + ".singlevar.cov.txt";
   else
      filename = traitName + ".singlevar.cov.txt";

   SCOREcov = fopen(filename,"wt");
   PrintHeader(SCOREoutput,engine,trans);

   Matrix genotypeAll; //this matrix has the all marker genotype within a window
   genotypeAll.Dimension(0,0);

   IntArray markerInLD;

   Vector sigma2,transGeno;
   transGeno.Dimension(trans.persons);
   sigma2.Dimension(trans.persons);

   if(genoFromPed)
   {
      //TODO: this assumes all markers are sorted by chromosome and position
      StringArray name;
      name.AddTokens(ped.markerNames[0],":");

      String initial_chr = name[0];
      int initial_pos = atoi(name[1].c_str());
      //printf("initial: %s %d\n",initial_chr.c_str(),initial_pos);

      for(int m=0;m<ped.markerCount;m++)
      {
	 if(!GetGenotypeVector(trans,ped,m))
	    continue;

	 //output score statistics
	 if(FastFit::unrelated)
	 {
	    double numerator = 0.0;
	    double denominator = 1.0e-10;
	    double chisq,pvalue,effSize;
	    for(int i=0;i<trans.persons;i++)
	    {
	       numerator += engine.residuals_unrelated[i]*genotype[i]/engine.sigma2Hat;
	       denominator += genotype[i]*genotype[i]/engine.sigma2Hat;
	    }
	    if(denominator==0) chisq=0.0;
	    else chisq = numerator*numerator/denominator;
	    pvalue = chidist(chisq, 1);

	    //This is the alternative allele effect size
	    effSize = numerator/denominator;
	    //fprintf(SCOREoutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%g\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),effSize,pvalue);
	    fprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,numerator,effSize,pvalue);
	 }
	 else
	 {
	    //get transGeno
	    if(!trans.empKin)
	    {
	       transGeno.Zero();
	       int index=0;
	       for (int f = 0; f < ped.familyCount; f++)
		  if (trans.pPheno[f].Length() > 0)
		  {
		     int count = trans.pPheno[f].Length();
		     int block = index;
		     for(int i=0;i<count;i++)
		     {
			for(int c=0;c<count;c++)
			{
			   transGeno[index] += trans.transU[block+i][block+c]*genotype[block+c];
			}
			index++;
		     }
		  }
	    } else transGeno.Product(trans.transU,genotype);

	    //calculate score statistics and output here
	    double sigma_g2 = engine.sigma_g2Hat;
	    double sigma_e2 = engine.sigma_e2Hat;

	    if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
	    {
	       for(int i=0;i<trans.persons;i++)
		  sigma2[i] = trans.D[i];
	    }
	    else
	    {
	       for(int i=0;i<trans.persons;i++)
	       {
		  sigma2[i] =sigma_g2*trans.D[i]+sigma_e2;
	       }
	    }
	    double numerator = 0.0;
	    double denominator = 1.0e-10;
	    double fixedEffect,chisq,pvalue,effSize;
	    for(int i=0;i<trans.persons;i++)
	    {
	       fixedEffect=trans.UX[i].InnerProduct(engine.betaHat);
	       numerator += (trans.UY[i]-fixedEffect)*transGeno[i]/sigma2[i];
	       denominator += transGeno[i]*transGeno[i]/sigma2[i];
	    }
	    if(denominator==0) chisq=0.0;
	    else chisq = numerator*numerator/denominator;
	    //printf("numerator and denominator are: %g %g\n",numerator, denominator);
	    pvalue = chidist(chisq, 1);
	    //This is the alternative allele effect size
	    effSize = numerator/denominator;
	    //fprintf(SCOREoutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%g\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),effSize,pvalue);
	    fprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,numerator,effSize,pvalue);
	 }

	 if(chr != initial_chr)
	 {
	    //output cov for the rest of the markers saved in genotypeAll
	    int markers = genotypeAll.cols;
	    for(int i=0;i<markers;i++)
	    {
	       initial_pos = markerInLD[0];
	       //       printf("current position: %d; initial position: %d,markerInLD length %d\t",pos,initial_pos,markerInLD.Length());
	       fprintf(SCOREcov,"%s\t%d\t",initial_chr.c_str(),initial_pos);
	       for(int j=0;j<markerInLD.Length();j++)
	       {
		  fprintf(SCOREcov,"%d,",markerInLD[j]);
	       }
	       fprintf(SCOREcov,"\t");
	       for(int m=0;m<genotypeAll.cols;m++)
	       {
		  double cov = CalculateCov(engine,trans,ped,genotypeAll,sigma2,m);
		  fprintf(SCOREcov,"%g,",cov);
	       }
	       fprintf(SCOREcov,"\n");

	       //reorganize the trackers
	       if(markerInLD.Length()>0)
	       {
		  markerInLD.Delete(0);
		  genotypeAll.DeleteColumn(0);
	       }
	       //    printf("markerInID length after %d\n",markerInLD.Length());
	    }

	    //Start a new sequence
	    initial_chr = chr;
	    initial_pos = pos;
	    markerInLD.Clear();
	    genotypeAll.Dimension(0,0);
	 }
	 else
	 {
	    while(pos - initial_pos > window && markerInLD.Length() >0 )
	    {
	       //output cov for marker at initial_pos
	       initial_pos = markerInLD[0];
	       //      printf("current position: %d; initial position: %d,markerInLD length %d\t",pos,initial_pos,markerInLD.Length());
	       fprintf(SCOREcov,"%s\t%d\t",chr.c_str(),initial_pos);
	       for(int i=0;i<markerInLD.Length();i++)
	       {
		  fprintf(SCOREcov,"%d,",markerInLD[i]);
	       }
	       fprintf(SCOREcov,"\t");
	       for(int m=0;m<genotypeAll.cols;m++)
	       {
		  double cov = CalculateCov(engine,trans,ped,genotypeAll,sigma2,m);
		  fprintf(SCOREcov,"%g,",cov);
	       }
	       fprintf(SCOREcov,"\n");

	       markerInLD.Delete(0);
	       genotypeAll.DeleteColumn(0);
	    }
	    //	    printf("markerInID length after %d\n",markerInLD.Length());
	 }
	 //if not, then update current sequence  
	 markerInLD.Push(pos);
	 initial_pos = markerInLD[0];
	 genotypeAll.GrowTo(trans.persons,genotypeAll.cols+1,0.0);
	 for(int i=0;i<trans.persons;i++)
	 {
	    if(FastFit::unrelated)
	       genotypeAll[i][genotypeAll.cols-1] = genotype[i];
	    else
	       genotypeAll[i][genotypeAll.cols-1] = transGeno[i];
	 }
      }
      //output cov for the rest of the markers saved in genotypeAll
      int markers = genotypeAll.cols;
      for(int i=0;i<markers;i++)
      {
	 initial_pos = markerInLD[0];
	 fprintf(SCOREcov,"%s\t%d\t",chr.c_str(),initial_pos);
	 for(int j=0;j<markerInLD.Length();j++)
	 {
	    fprintf(SCOREcov,"%d,",markerInLD[j]);
	 }
	 fprintf(SCOREcov,"\t");
	 for(int m=0;m<genotypeAll.cols;m++)
	 {
	    double cov = CalculateCov(engine,trans,ped,genotypeAll,sigma2,m);
	    fprintf(SCOREcov,"%g,",cov);
	 }
	 fprintf(SCOREcov,"\n");

	 //reorganize the trackers
	 if(markerInLD.Length()>0)
	 {
	    markerInLD.Delete(0);
	    genotypeAll.DeleteColumn(0);
	 }
      }
   }

   if(genoFromVCF)
   {
      StringArray chromosome;
      chromosome.Push("1");chromosome.Push("2");chromosome.Push("3");chromosome.Push("4");chromosome.Push("5");chromosome.Push("6");chromosome.Push("7");chromosome.Push("8");chromosome.Push("9");chromosome.Push("10");chromosome.Push("11");chromosome.Push("12");chromosome.Push("13");chromosome.Push("14");chromosome.Push("15");chromosome.Push("16");chromosome.Push("17");chromosome.Push("18");chromosome.Push("19");chromosome.Push("20");chromosome.Push("21");chromosome.Push("22");chromosome.Push("X");chromosome.Push("Y");chromosome.Push("M");

      //loop through chromosomes by reading tabix data
      for(int i=0;i<chromosome.Length();i++)
      {
	 reader.open(vcfInput,header);
	 reader.readVcfIndex();
	 reader.setReadSection(chromosome[i].c_str());
	 int initial_pos = 0;
	 int marker_count = 0;

	 while(reader.readRecord(record))
	 {
	    if(!GetGenotypeVector(trans,ped,0))
	       continue;

	    marker_count++;

	    if(marker_count==1) 
	       initial_pos = pos;


	    //Output score statistics here
	    if(FastFit::unrelated)
	    {
	       double numerator = 0.0;
	       double denominator = 1.0e-10;
	       double chisq,pvalue,effSize;
	       for(int i=0;i<trans.persons;i++)
	       {
		  //fixedEffect=trans.X[i].InnerProduct(engine.betaHat);
		  numerator += engine.residuals_unrelated[i]*genotype[i]/engine.sigma2Hat;
		  denominator += genotype[i]*genotype[i]/engine.sigma2Hat;
	       }
	       if(denominator==0) chisq=0.0;
	       else chisq = numerator*numerator/denominator;
	       //printf("numerator and denominator are: %g %g\n",numerator, denominator);
	       pvalue = chidist(chisq, 1);
	       //This is the alternative allele effect size
	       effSize = numerator/denominator;
	       //fprintf(SCOREoutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%g\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),effSize,pvalue);
	       fprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,numerator,effSize,pvalue);
	    }
	    else
	    {
	       //get transGeno
	       if(!trans.empKin)
	       {
		  transGeno.Zero();
		  int index=0;
		  for (int f = 0; f < ped.familyCount; f++)
		     if (trans.pPheno[f].Length() > 0)
		     {
			int count = trans.pPheno[f].Length();
			int block = index;
			for(int i=0;i<count;i++)
			{
			   for(int c=0;c<count;c++)
			   {
			      transGeno[index] += trans.transU[block+i][block+c]*genotype[block+c];
			   }
			   index++;
			}
		     }
	       } else transGeno.Product(trans.transU,genotype);

	       //calculate score statistics and output here
	       double sigma_g2 = engine.sigma_g2Hat;
	       double sigma_e2 = engine.sigma_e2Hat;
	       //printf("%g %g\n",sigma_g2,sigma_e2);
	       if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
	       {
		  for(int i=0;i<trans.persons;i++)
		     sigma2[i] = trans.D[i];
	       }
	       else
	       {
		  for(int i=0;i<trans.persons;i++)
		  {
		     sigma2[i] =sigma_g2*trans.D[i]+sigma_e2;
		  }
	       }
	       double numerator = 0.0;
	       double denominator = 1.0e-10;
	       double fixedEffect,chisq,pvalue,effSize;
	       for(int i=0;i<trans.persons;i++)
	       {
		  fixedEffect=trans.UX[i].InnerProduct(engine.betaHat);
		  numerator += (trans.UY[i]-fixedEffect)*transGeno[i]/sigma2[i];
		  denominator += transGeno[i]*transGeno[i]/sigma2[i];
	       }
	       if(denominator==0) chisq=0.0;
	       else chisq = numerator*numerator/denominator;
	       //printf("numerator and denominator are: %g %g\n",numerator, denominator);
	       pvalue = chidist(chisq, 1);
	       //This is the alternative allele effect size
	       effSize = numerator/denominator;
	       //   fprintf(SCOREoutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%g\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),effSize,pvalue);
	       fprintf(SCOREoutput,"%s\t%d\t%s\t%s\t%d\t%g\t%g\t%g\t%g\t%g\t%g\n",chr.c_str(),pos,refAllele.c_str(),rareAllele.c_str(),trans.persons,founderMaf,markerMaf,mac,numerator,effSize,pvalue);
	    }

	    //check if position has passed the window
	    while(pos - initial_pos > window && markerInLD.Length() >0 )
	    {
	       //output cov for marker at initial_pos
	       fprintf(SCOREcov,"%s\t%d\t",chr.c_str(),initial_pos);
	       for(int i=0;i<markerInLD.Length();i++)
	       {
		  fprintf(SCOREcov,"%d,",markerInLD[i]);
	       }
	       fprintf(SCOREcov,"\t");
	       for(int m=0;m<genotypeAll.cols;m++)
	       {
		  double cov = CalculateCov(engine,trans,ped,genotypeAll,sigma2,m);
		  fprintf(SCOREcov,"%g,",cov);
	       }
	       fprintf(SCOREcov,"\n");

	       //reorganize the trackers
	       markerInLD.Delete(0);
	       initial_pos = markerInLD[0];
	       genotypeAll.DeleteColumn(0);
	    }

	    //push the new marker in
	    markerInLD.Push(pos);
	    initial_pos = markerInLD[0];
	    genotypeAll.GrowTo(trans.persons,genotypeAll.cols+1,0.0);
	    for(int i=0;i<trans.persons;i++)
	    {
	       if(FastFit::unrelated)
		  genotypeAll[i][genotypeAll.cols-1] = genotype[i];
	       else
		  genotypeAll[i][genotypeAll.cols-1] = transGeno[i];
	    }
	 }
	 //output the LD matrix for the rest of the markers in genotypeAll
	 int markers = genotypeAll.cols;
	 for(int i=0;i<markers;i++)
	 {
	    fprintf(SCOREcov,"%s\t%d\t",chr.c_str(),initial_pos);
	    for(int j=0;j<markerInLD.Length();j++)
	    {
	       fprintf(SCOREcov,"%d,",markerInLD[j]);
	    }
	    fprintf(SCOREcov,"\t");
	    for(int m=0;m<genotypeAll.cols;m++)
	    {
	       double cov = CalculateCov(engine,trans,ped,genotypeAll,sigma2,m);
	       fprintf(SCOREcov,"%g,",cov);
	    }
	    fprintf(SCOREcov,"\n");

	    //reorganize the trackers
	    markerInLD.Delete(0);
	    initial_pos = markerInLD[0];
	    genotypeAll.DeleteColumn(0);
	 }
	 reader.close();
      }
   }
   fclose(SCOREoutput);
   fclose(SCOREcov);
}

bool PreMeta::GetGenotypeVectorVCFDosage(FastTransform & trans, Pedigree & ped)
{
   bool status = false;
   rareAllele = record.getAltStr();
   refAllele = record.getRefStr();
   chr = record.getChromStr();
   pos = record.get1BasedPosition();

   //estimate allele frequencies here
   //printf("start estimating allele frequencies.\n");

   int numSamples = header.getNumSamples();

   int n=0,nf=0;
   double fmaf = 0.0, maf=0.0;
   VcfRecordGenotype & genoInfo = record.getGenotypeInfo();
   for(int s=0;s<numSamples;s++)
   {
      const std::string * geno = genoInfo.getString("DS",s);
      const char * sample = header.getSampleName(s);

      if(geno==NULL)
      {
	 error("Error: Can not find DS field for sample %d ...",s);
      }
      if(*geno == ".")
      {
	 continue;
      }
      float dose = atof((*geno).c_str());
      //judge if this is a founder

      int founder = trans.foundersHash.Integer(sample);
      if(founder != -1)
      {
	 fmaf += dose;
	 nf++;
      }
      //CalculateCov overall maf
      maf += dose;
      n++;
   }

   //calculate final maf
   if(nf!=0)
      fmaf /= 2.0*nf;
   maf /= 2.0*n;
   if(nf==0.0 && maf >0.0)
      fmaf=maf;

   founderMaf = fmaf;
   markerMaf = maf;

   if(FounderFreq && fmaf==0.0)
   {
      return status;
   }
   else
   {
      if(maf==0.0)
      {
	 return status;
      }
   }

   bool XStatus = false;
   if(AutoFit::fitX)
   {
      if(chr==xLabel)
	 XStatus=true;
   }

   //printf("Now working on marker %d.\n",markerIdx);
   genotype.Dimension(trans.persons);
   for(int i=0;i<trans.persons;i++)
      genotype[i]=-1.0;

   mac=0.0;

   //Loop through samples according to trans.pPheno 
   //read in dosage into genotype vector
   int idx=0;
   int n_ = 0;
   double mean = 0.0;
   for (int f = 0; f < ped.familyCount; f++)
   {
      for(int j=0;j<trans.pPheno[f].Length();j++)
      {
	 String sample = ped[trans.pPheno[f][j]].pid;

	 //get the sample number by sample ID from hash
	 int s = trans.sampleVCFIDHash.Integer(sample);
	 if(s!=-1)
	 {
	    const std::string * geno = genoInfo.getString("DS",s);
	    if(strcmp((*geno).c_str(),".")>0)
	    {
	       float dose = atof((*geno).c_str());
	       genotype[idx] = dose;
	       if(XStatus)
	       {
		  if(dose>0)
		     genotype[idx] = 2;
	       }
	       n_++;
	       mean += dose;
	       mac += dose;
	       status = true;
	    }
	 }
	 idx++;
      }
   }
   mean /= n_;
   for(int i=0;i<trans.persons;i++)
   {
      if(genotype[i]==-1.0)
      {
	 genotype[i] = 0.0;
      }
      else
      {
	 genotype[i] -= mean;
      }
   }
   return status;
}

bool PreMeta::GetGenotypeVectorVCF(FastTransform & trans, Pedigree & ped)
{
   bool status = false;

   rareAllele = record.getAltStr();
   refAllele = record.getRefStr();
   chr = record.getChromStr();
   pos = record.get1BasedPosition();

   //estimate allele frequencies here
   int numSamples = header.getNumSamples();
   double fmaf=0.0,maf=0.0;
   std::vector<int> fcounts;
   std::vector<int> counts;

   for(int i = 0; i < numSamples; i++)
   {
      const char * sample = header.getSampleName(i);
      int numGTs = record.getNumGTs(i);
      Vector allele;
      for(int j = 0; j < numGTs; j++)
      {
	 allele.Push(record.getGT(i,j));
      }
      int bad =0;
      for(int m=0;m<allele.Length();m++)
      {
	 bad += (allele[m]==VcfGenotypeSample::MISSING_GT);
      }
      if(bad>0)
      {
	 //there is at least one allele missing, so skip this sample
	 continue;
      }

      //judge if this is a founder
      int founder = trans.foundersHash.Integer(sample);

      for(int m=0;m<allele.Length();m++)
      {
	 if(allele[m] >= counts.size())
	 {
	    counts.resize(allele[m]+1);
	 }
	 ++counts[allele[m]];
	 if(founder != -1)
	 {
	    if(allele[m] >= fcounts.size())
	    {
	       fcounts.resize(allele[m]+1);
	    }
	    ++fcounts[allele[m]];
	 }
      }
      //Only consider diallelic markers
      if(counts.size()>2)
      {
	 fmaf = -1.0;
	 maf = -1.0;
      }
   }

   //if mornomorphic, then skip this marker
   if(counts.size()==1 || fcounts.size()==1)
   {
      return status;
   }

   //if not biallelic, then skip this marker
   if(fmaf==-1.0 ||maf==-1.0)
   {
      return status;
   }

   //CalculateCov maf
   double sum = 0.0;
   for(int i=0;i<counts.size();i++)
   {
      sum+= counts[i];
   }
   maf = counts[1]/sum;
   sum = 0.0;
   for(int i=0;i<fcounts.size();i++)
   {
      sum+= fcounts[i];
   }
   if(sum>0)
      fmaf = fcounts[1]/sum;
   if(fmaf==0.0 && maf >0.0)
      fmaf=maf;

   founderMaf = fmaf;
   markerMaf = maf;


   bool XStatus = false;
   if(AutoFit::fitX)
   {
      if(chr==xLabel)
	 XStatus=true;
   }

   //CalculateCov and output score statistic here
   //   printf("Now working on marker %d.\n",markerIdx);
   genotype.Dimension(trans.persons);
   for(int i=0;i<trans.persons;i++)
      genotype[i] = -1.0;

   mac=0.0;

   //Loop through samples according to trans.pPheno 
   //read in dosage into genotype vector
   int idx=0;
   int n_=0;
   double mean = 0.0;
   for (int f = 0; f < ped.familyCount; f++)
   {
      for(int j=0;j<trans.pPheno[f].Length();j++)
      {
	 String sample = ped[trans.pPheno[f][j]].pid.c_str();

	 //get the sample number by sample ID from hash
	 int s = trans.sampleVCFIDHash.Integer(sample);
	 if(s!=-1)
	 {
	    int numGTs = record.getNumGTs(s);
	    Vector allele;

	    for(int j = 0; j < numGTs; j++)
	    {
	       allele.Push(record.getGT(s,j));
	    }
	    int bad =0;
	    for(int m=0;m<allele.Length();m++)
	    {
	       bad += (allele[m]==VcfGenotypeSample::MISSING_GT);
	    }
	    //if any allele is missing then use 2*maf to impute
	    if(bad==0)
	    {
	       status = true;
	       int sum=0;
	       for(int i=0;i<allele.Length();i++)
		  sum+=allele[i];
	       genotype[idx] = sum;
	       if(XStatus)
	       {
		  if(sum>0)
		     genotype[idx] = 2;
	       }
	       mean += sum;
	       n_++;
	       mac += sum;
	    }
	 }
	 idx++;
      }
   }
   //Adjust genotype here
   mean /= n_;
   for(int i=0;i<trans.persons;i++)
   {
      if(genotype[i]==-1.0)
      {
	 genotype[i] = 0.0;
      }
      else
      {
	 genotype[i] -= mean;
      }
   }
   return status;
}

bool PreMeta::GetGenotypeVector(FastTransform & trans, Pedigree & ped,int markerid)
{
   bool status;
   if(dosage)
   {
      status = GetGenotypeVectorVCFDosage(trans,ped);
   }
   if(genoFromPed)
   {
      status = GetGenotypeVectorPED(trans,ped,markerid);
   }
   if(genoFromVCF && !dosage)
   {
      status = GetGenotypeVectorVCF(trans,ped);
   }
   return status;
}

bool PreMeta::GetGenotypeVectorPED(FastTransform & trans, Pedigree & ped,int markerid)
{
   bool status = false;

   StringArray markerName;
   markerName.AddTokens(ped.markerNames[markerid],":");
   chr = markerName[0];
   pos = atoi(markerName[1].c_str());

   int alleleCount = CountAlleles(ped, markerid);
   //if no one is genotyped, mornomorphic, or marker is not diallelic
   if(alleleCount<=1 || alleleCount>2)
      return status;
   //Estimate allele frequency here
   Vector freq; int minor;
   markerMaf = 0.0;
   bool tmpF = EstimateFrequencies(ped,markerid,FREQ_ALL);
   freq.Copy(ped.GetMarkerInfo(markerid)->freq);
   if(freq[1] < freq[2]) { minor=1; markerMaf = freq[1]; }
   else
   {
      markerMaf = freq[2];
      minor = 2;
   }
   int major;

   if(minor==1) major=2;
   else major =1;
   rareAllele = ped.GetMarkerInfo(markerid)->GetAlleleLabel(minor);
   refAllele = ped.GetMarkerInfo(markerid)->GetAlleleLabel(major);

   //Estimate founder allele frequencies here
   tmpF = EstimateFrequencies(ped,markerid,FREQ_FOUNDERS);
   freq.Copy(ped.GetMarkerInfo(markerid)->freq);
   founderMaf = freq[1];
   if(founderMaf>0.5) founderMaf = 1.0-founderMaf;

   int idx=0;
   mac =0.0;
   bool XStatus = false;
   if(AutoFit::fitX)
   {
      String markerName = ped.markerNames[markerid];
      int found = ped.markerNames[markerid].FastFind(xLabel,0);
      StringArray tmp;
      tmp.AddTokens(markerName,":");
      if(found!=-1)
      {
	 if(atoi(tmp[1].c_str()) > PreMeta::Xstart && atoi(tmp[1].c_str())< PreMeta::Xend)
	 {
	    XStatus=true;
	 }
      }
   }
   genotype.Dimension(trans.persons);
   genotype.Zero();
   for(int f=0;f<ped.familyCount;f++)
   {
      for(int p=0;p<trans.pPheno[f].Length();p++)
      {
	 if(!ped[trans.pPheno[f][p]].isGenotyped(markerid))
	 {
	    genotype[idx]=0.0;
	    //genotype[idx]=2.0*markerMaf;   
	 }
	 else if(ped[trans.pPheno[f][p]].markers[markerid].isHeterozygous())
	 {
	    if(XStatus)
	    {
	       if(ped[trans.pPheno[f][p]].sex==2)
	       {
		  genotype[idx]=1.0 - 2.0*markerMaf;
		  status = true;
		  mac++;
	       }
	       if(ped[trans.pPheno[f][p]].sex==1)
	       {
		  genotype[idx]=2.0 - 2.0*markerMaf;
		  status = true;
		  mac++;
	       }
	    }
	    else
	    {
	       genotype[idx]=1.0 - 2.0*markerMaf;
	       status = true;
	       mac++;
	    }
	 }
	 else if(ped[trans.pPheno[f][p]].markers[markerid].isHomozygous() && ped[trans.pPheno[f][p]].markers[markerid].hasAllele(minor))
	 {
	    genotype[idx]=2.0 - 2.0*markerMaf;
	    status = true;
	    mac++;
	    mac++;
	 }
	 else
	 {
	    genotype[idx]=0.0-2.0*markerMaf;
	 }
	 idx++;
      }
   }
   return status;
}

void PreMeta::PrintHeader(FILE * output,FastFit & engine,FastTransform & trans)
{
   double sigma_g2 = engine.sigma_g2Hat;
   double h2 = 100.0*sigma_g2/trans.traitVar;
   double h2X = 100.0*engine.sigma_gXHat/trans.traitVar;
   double sharedEnv =100.0*engine.sigma_s2Hat/trans.traitVar;
   if(!AutoFit::fitSharedEnvironment && !AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2);
   }
   else if(AutoFit::fitSharedEnvironment && AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f, Xheritability = %6.2f, SharedEnv = %6.2f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,h2X,sharedEnv);
   }
   else if(AutoFit::fitSharedEnvironment)
   {
      fprintf(output,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f, SharedEnv = %6.2f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,sharedEnv);
   }
   else if(AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f, Xheritability = %6.2f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,h2X);
   }
   fprintf(output,"===================================================================\n");
   //fprintf(output,"\nCHR\tPOS\t#FOUNDERS\tFOUNDER_AF\tALL_AF\tINFORMATIVE_AC\tREF_ALLELE\tALT_ALLELE\tALT_ALLELE_EFFSIZE\tPVALUE\n");
   fprintf(output,"\nCHR\tPOS\tREF_ALLELE\tALT_ALLELE\tINFORMATIVE_N\tFOUNDER_AF\tALL_AF\tINFORMATIVE_AC\tSTAT\tALT_ALLELE_EFFSIZE\tPVALUE\n");
}

